/*--------------------------------------------------------------------------
debug.c:
Debugger interface.
A lot of ANSI Escapes here :)
--------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include "pic.h"

extern word eetim;

//*******************************************************************
// Dissasembler functions:

char *regname(byte r)
{
        static char buf[32];
        const char *reg[12]={ "INDF","TMR0","PCL","STATUS",
                             "FSR","PORTA","PORTB","?7?",
                             "EEDATA","EEADR","PCLATCH",
                             "INTCON" };
        if (r<12) sprintf(buf,"%s",reg[r]);
        else sprintf(buf,"f_0x%02X",r);
        return buf;
}

char *f84_dis(word ireg)
{
        static char buf[64];
        word lk;
        byte f,d,b,k,tmp;
        int i,j,n;

        // Extract operands fields
        lk=ireg&0x7ff; k=ireg&0xff;
        f=ireg&0x7f; d=ireg&0x80;
        b=(ireg>>7)&7;
        switch(ireg>>11) {
                case 4: // CALL
                    sprintf(buf,"CALL    0x%03X",lk); break;
                case 5: // GOTO
                    sprintf(buf,"GOTO    0x%03X",lk); break;
                default:
                    switch(ireg>>10) {
                    case 0xc:   // MOVLW
                        sprintf(buf,"MOVLW   0x%02X",k); break;
                    case 0xd:   // RETLW
                        sprintf(buf,"RETLW   0x%02X",k); break;
                        break;
                    case 4:     // BCF
                        sprintf(buf,"BCF     %s,%d",regname(f),b); break;
                    case 5:     // BSF
                        sprintf(buf,"BSF     %s,%d",regname(f),b); break;
                    case 6:     // BTFSC
                        sprintf(buf,"BTFSC   %s,%d",regname(f),b); break;
                    case 7:     // BTFSS
                        sprintf(buf,"BTFSS   %s,%d",regname(f),b); break;
                    default:
                        switch(ireg>>9) {
                        case 0x1f: // ADDLW
                            sprintf(buf,"ADDLW   0x%02X",k); break;
                        case 0x1e: // SUBLW
                            sprintf(buf,"SUBLW   0x%02X",k); break;
                        default:
                            switch(ireg>>8){
                            case 0x7: // ADDWF
                                sprintf(buf,"ADDWF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x5: // ANDWF
                                sprintf(buf,"ANDWF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x9: // COMF
                                sprintf(buf,"COMF    %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x3: // DECF
                                sprintf(buf,"DECF    %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xb: // DECFSZ
                                sprintf(buf,"DECFSZ  %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xa: // INCF
                                sprintf(buf,"INCF    %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xf: // INCFSZ
                                sprintf(buf,"INCFSZ  %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x4: // IORWF
                                sprintf(buf,"IORWF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x8: // MOVF
                                sprintf(buf,"MOVF    %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xd: // RLF
                                sprintf(buf,"RLF     %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xc: // RRF
                                sprintf(buf,"RRF     %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x2: // SUBWF
                                sprintf(buf,"SUBWF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0xe: // SWAPF
                                sprintf(buf,"SWAPF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x6: // XORWF
                                sprintf(buf,"XORWF   %s,%c",regname(f),
                                        (d)?'f':'w'); break;
                            case 0x39: // ANDLW
                                sprintf(buf,"ANDLW   0x%02X",k); break;
                            case 0x38: // IORLW
                                sprintf(buf,"IORLW   0x%02X",k); break;
                            case 0x3a: // XORLW
                                sprintf(buf,"XORLW   0x%02X",k); break;
                            default:
                                switch(ireg>>7) {
                                case 0x3: // CLRF
                                    sprintf(buf,"CLRF    %s",regname(f));
                                    break;
                                case 0x2: // CLRW
                                    sprintf(buf,"CLRW"); break;
                                case 0x1: // MOVWF
                                    sprintf(buf,"MOVWF   %s",regname(f));
                                    break;
                                default:
                                    switch(ireg) {
                                    case 0x64: // CLRWDT
                                        sprintf(buf,"CLRWDT"); break;
                                    case 0x9: // RETFIE
                                        sprintf(buf,"RETFIE"); break;
                                    case 0x8: // RETURN
                                        sprintf(buf,"RETURN"); break;
                                    case 0x63: // SLEEP
                                        sprintf(buf,"SLEEP"); break;
                                    case 0x65: // TRIS PORTA
                                        sprintf(buf,"TRIS    PORTA"); break;
                                    case 0x66: // TRIS PORTB
                                        sprintf(buf,"TRIS    PORTB"); break;
                                    case 0x62: // OPTION
                                        sprintf(buf,"OPTION"); break;
                                    default:
                                        sprintf(buf,"NOP"); break;
                                    }
                                }
                            }
                        }
                    }
        }
        return buf;
}

//---------------------------------------------------------------------------
// Special function registers are printed bit by bit. Here are the bit names:
//---------------------------------------------------------------------------
const char *status_names[8]={"C ","DC ","Z ","/PD ","/TO ","RP0 ","",""};
const char *porta_names[8]={"RA0 ","RA1 ","RA2 ","RA3 ","RA4/T0CKI ",
                                "","",""};
const char *portb_names[8]={"RB0/INT ","RB1 ","RB2 ","RB3 ","RB4 ","RB5 ",
                                "RB6 ","RB7 "};
const char *intcon_names[8]={"RBIF ","INTF ","T0IF ","RBIE ","INTE ","T0IE ",
                                "EEIE ","GIE "};
const char *option_names[8]={"PS0 ","PS1 ","PS2 ","PSA ","T0SE ","T0CS ",
                                "INTEDG ","/RBPU "};
const char *eecon1_names[8]={"RD ","WR ","WREN ","WRERR ","EEIF ","","",""};

// This function prints a register using a reverse video for 1's
void prt_reg(byte val, const char *bitnames[],int color)
{
        int i;
        for (i=7;i>=0;i--) {
                if (val&0x80) printf("\033[01;3%dm\033[7m%s\033[m",
                                color,bitnames[i]);
                else printf("\033[01;3%dm%s\033[m",color,bitnames[i]);
                val<<=1;
        }
}
// This function prints the status of PortA or PortB. The color is changed
// depending on the direction of the port bit.
void prt_port(byte port,core *pcore,int colorin,int colorout)
{
        int i,val,valt;
        char **p;
        if (port) {
                //val=pcore->portb;
                val=(portb_rd(pcore) & pcore->trisb) |
                    (pcore->portb & (~pcore->trisb));
                valt=pcore->trisb;
                p=(char **)portb_names;
        } else {
                //val=pcore->porta;
                val=(porta_rd(pcore) & pcore->trisa) |
                    (pcore->porta & (~pcore->trisa));
                valt=pcore->trisa;
                p=(char **)porta_names;
        }
        for (i=7;i>=0;i--) {
                if (val&0x80) printf("\033[01;3%dm\033[7m%s\033[m",
                                (valt&0x80)?colorin:colorout,p[i]);
                else printf("\033[01;3%dm%s\033[m",
                        (valt&0x80)?colorin:colorout,p[i]);
                val<<=1; valt<<=1;
        }
}

// Function to print the tristare registers. The problem here is that
// the port names do not have a fixed name length.
void prt_tris(byte val,byte port,int color)
{
        int i,n;
        printf("\033[01;3%dm",color);
        if (port) n=7;
        else {n=4; val<<=3;}
        for (i=n;i>=0;i--) {
                if (port==1 && i==0) printf("  ");
                if (port==0 && i==4) printf("    ");
                printf("%s",(val&0x80)?"IN  ":"OUT ");
                if (port==0 && i==4) printf("  ");
                val<<=1;
        }
        printf("\033[m");
}

void gotoxy(int x,int y)
{
         printf("\033[%d;%dH",y+1,x+1);
}

//----------------------------------------------------------------------
// Debug interface

int debug_status=STEP;  // Debugging status (STEP,TRACE,RUNNING)
unsigned long tbrk=0;   // Time breakpoint
word brk=0xffff;        // Breakpoint

void debug(core *pcore,word *prog)
{
    int i,j,k,pc;
    static byte tf[68];

    for(;;) {
        // Dump current state:
        if (debug_status==STEP) printf("\033[2J");
        printf("\033[H\033[m");        // Clear & Home
        for (i=0;i<8;i++) {
                printf((i==0)?"STACK: ":"       ");
                if (i==pcore->sp) printf("\033[01;34m");
                printf("%03X\033[m\n",pcore->pc[i]);
        }
        printf("\n\n");
        pc=pcore->pc[pcore->sp];
        // A few instructions before PC
        for (i=-4;i<0;i++) {
                j=(pc+i)&0x3ff;
                printf("                    \r");
                printf("%03X %s\n",j,f84_dis(prog[j]));
        }
        // Current in-decoding instruction
        printf("                    \r");
        printf("\033[01;31m    %s\033[m\n",f84_dis(pcore->ir));
        // Current PC-fetching instruction
        printf("                    \r");
        printf("\033[01;34m%03X %s\033[m\n",pc,f84_dis(prog[pc]));
        // A few instructions after PC
        for (i=1;i<8;i++) {
                printf("                    \r");
                printf("%03X %s\n",pc+i,f84_dis(prog[pc+i]));
        }

        // REGISTERS:
        gotoxy(12,0); printf(    "\033[1m     W: %02X (%03d)",pcore->w,pcore->w);
        gotoxy(12,1); printf("\033[01;30mPresc.: %02X (%03d)",
                        pcore->prescaler,pcore->prescaler);
        gotoxy(12,2); printf("\033[01;30m  TMR0: %02X (%03d)",
                        pcore->tmr0,pcore->tmr0);
        gotoxy(12,3); printf("\033[01;31mSTATUS: ");
        prt_reg(pcore->status,status_names,1);
        gotoxy(12,4); printf("\033[01;32m   FSR: %02X (%03d)",
                        pcore->fsr,pcore->fsr);
        gotoxy(12,5); printf("\033[01;34m PORTA: ");
        prt_port(0,pcore,4,1);
        gotoxy(12,6); printf("\033[01;34m TRISA: ");
        prt_tris(pcore->trisa,0,4);
        gotoxy(12,7); printf("\033[01;34m PORTB: ");
        prt_port(1,pcore,4,1);
        gotoxy(12,8); printf("\033[01;34m TRISB: ");
        prt_tris(pcore->trisb,1,4);
        gotoxy(32,0); printf("\033[01;31mINTCON: ");
        prt_reg(pcore->intcon,intcon_names,1);
        gotoxy(32,1); printf("\033[01;32mOPTION: ");
        prt_reg(pcore->option,option_names,2);
        gotoxy(48,2); printf("\033[01;35mEECON1: ");
        prt_reg(pcore->eecon1,eecon1_names,5);
        gotoxy(48,3); printf("\033[01;35m EEADR: %02X",pcore->eeadr);
        gotoxy(48,4);
        printf("\033[01;35mEEDATA rd: %02X (%03d),wr: %02X (%03d)\033[m",
                pcore->eeprom[pcore->eeadr],pcore->eeprom[pcore->eeadr],
                pcore->eedatawr,pcore->eedatawr);
        gotoxy(48,5);
        if (eetim) {
            printf("\033[01;35mWriting: %02d%%\033[m",
                99-(int)eetim*1000/(4*pcore->freq/100));
        } else printf("            ");

        if (pcore->config&0x4) {
            gotoxy(58,6);
            printf("\033[01;30m WDT: %02d%%",
                pcore->wdt*100/18/(pcore->freq/1000));
            if(pcore->option&8) printf(" Presc.: %02X",pcore->prescaler);
            else printf("             ");
        }
        gotoxy(58,7); printf("\033[01;30mCLKs: (%d)",pcore->Clocks);
        gotoxy(58,8); printf("\033[01;30m BRK: %03X\033[m",brk);
        gotoxy(0,9);
        if(pcore->sleep) printf("\033[01;31m---SLEEPING---\033[m");
        else printf("              ");
        // Dump general purpose registers
        gotoxy(24,10); printf("\033[1mGen. Purpose Regs.:\033[m");
        gotoxy(28,11);
        for (i=0;i<16;i++) {
                if (i==8) printf("|");
                printf("|%X ",i);
        } printf("|");
        for (j=0;j<5;j++) {
                gotoxy(24,12+j);
                printf("%02X - ",j*16);
                for (i=0;i<16;i++) {
                        k=j*16+i-12;
                        if (k==68) break;
                        if (i==8) printf(" ");
                        if (k<0) printf("   ");
                        else {
                            if (pcore->f[k]==tf[k])
                                printf("%02X ",pcore->f[k]);
                            else printf("\033[01;32m\033[7m%02X\033[m ",
                                        pcore->f[k]);
                            tf[k]=pcore->f[k];
                        }
                }
        }

        // Dump Eeprom
        gotoxy(24,17); printf("\033[1mEEPROM:\033[m");
        for (j=0;j<4;j++) {
                gotoxy(24,18+j);
                printf("%02X - ",j*16);
                for (i=0;i<16;i++) {
                        k=j*16+i;
                        if (i==8) printf(" ");
                        if (k==pcore->eeadr)
                           printf("\033[01;35m\033[7m%02X\033[m ",pcore->eeprom[k]);
                        else printf("%02X ",pcore->eeprom[k]);
                }
        }
        fflush(stdout);
        if (debug_status==TRACE) usleep(50000);
        if (debug_status!=STEP) return;
        // Commands:
        gotoxy(0,24); printf("\033[1mS\033[mtep \033[1mN\033[mext ");
        printf("\033[1mG\033[mo \033[1mT\033[mrace \033[1mR\033[meset ");
        printf("\033[1mB\033[mreakpoint \033[1mC\033[mycles ");
        printf("\033[1mL\033[mogic_An. ");
        printf("\033[1mQ\033[muit");
        fflush(stdout);
        k=0;
        do {
          switch(getchar()) {
        case 's': return;
        case 'n': brk=pc+1; debug_status=RUNNING; return;
        case 'g': debug_status=RUNNING;
                  tbrk=0;
go_cmd:           printf("\r\033[J--\033[01;31mRunning\033[m-- press \033[1mCtrl-C\033[m to interrupt");
                  fflush(stdout);
                  return;
        case 't': debug_status=TRACE;
                  printf("\r\033[J--\033[01;31mTracing\033[m-- press \033[1mCtrl-C\033[m to interrupt");
                  return;
        case 'r': f84_reset(pcore,0); debug_status=STEP; brk=0xffff; return;
        case 'q': setcanon();
                  printf("\033c");
                  exit(0);
        case 'b': printf("\r\033[JEnter Breakpoint Address (hex): ");
                  fflush(stdout);
                  setcanon();
                  scanf("%x",&i);
                  setnocanon();
                  brk=i;
                  printf("\033[2J");
                  k=1;
                  break;
        case 'c': printf("\r\033[JEnter number of Cycles to run (dec): ");
                  fflush(stdout);
                  setcanon();
                  scanf("%u",&i);
                  setnocanon();
                  tbrk=pcore->Clocks+i;
                  debug_status=RUNNING;
                  goto go_cmd;
        case 'l': la_show_events();
                  k=1;
                  break;
          }
        } while(!k);
    }
}
