#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pic.h"

// IO emulation routines
// DElta-SIgma ADC

#define R1 100e3
#define R2 100e3
#define R3 65e3
#define C  330e-9
#define VDD 5.0
#define VTH 1.4

float vc=0.0,vo=0.0,vn=0.0;
float otime=0.0;

void vupdate(core *pcore)
{
        float t,vin,dt,q1,q2,q3;
        static unsigned int ot=0,phase=0;
        int idt;

        idt=pcore->Clocks; idt-=ot;
        ot=pcore->Clocks;

        phase+=idt*4295*5;
        vin = VDD*(0.5+0.4*sin(M_PI*phase/2147483647));

        //t=pcore->Clocks;
        //t/=pcore->freq;
        //vin=VDD*t/40.0;

        if (vin>VDD) exit(0);
        dt=idt; dt/=pcore->freq;
        q1=dt*(vin-vc)/R1;
        q2=dt*(vo-vc)/R2;
        q3=dt*vc/R3;
        vc+=(q1+q2-q3)/C;
        otime=t;
}


// RD routines. Called very often to check timer & interrupt pins

byte porta_rd(core *pcore)
{
        static byte opa=0;
        byte pa;

        vupdate(pcore);
        pa=(vc>VTH+vn)?0x1f:0x17;

        if (pa!=opa) la_add_event(pcore);

        opa=pcore->porta;
        return pa;
}

byte portb_rd(core *pcore)
{
        static byte opb=0;
        byte pb;

        pb=0xff;

        if (pb!=opb) la_add_event(pcore);

        opb=pcore->portb;
        return pb;
}

// WR routines. Called only when program writes to ports

void porta_wr(core *pcore)
{
        static byte opa=0,otra=0xff;
        if (pcore->porta!=opa || pcore->trisa!=otra) la_add_event(pcore);
        vupdate(pcore);
        vo=((pcore->porta|pcore->trisa)&0x4)?VDD:0.0;
        vn=((rand()&0xff)-128)*0.002/128.;
        opa=pcore->porta; otra=pcore->trisa;
}

void portb_wr(core *pcore)
{
        static byte opb=0,otrb=0xff;

        if (pcore->portb!=opb || pcore->trisb!=otrb) la_add_event(pcore);
        opb=pcore->portb; otrb=pcore->trisb;
}

// This routine is called periodically every 10000 cycles
// Can be used to simulate slow hardware without performance
// penalty

void do_periodic_update() {}

// This routine is called on interrupts
// UART emulation
// (UART is synchonous with interrupts)

#define BMASK 0x80

void trace_interrupt(core *pcore)
{
        static byte st=0,ch;

        //printf("%c\n",(pcore->porta&0x4)?'1':'0');
        //return;

        if (!st) {
            if (((pcore->portb|pcore->trisb)&BMASK)==0) {ch=0; st++;}
        } else {
            ch>>=1; ch|=((pcore->portb|pcore->trisb)&BMASK)?0x80:0;
            st++;
            if (st==9) fputc(ch,stdout); fflush(stdout);
            if (st==10) st=0;
        }
}

