#include <stdio.h>
#include <stdlib.h>
#include "pic.h"

// Logic Analizer functions: Keep trace of ports

typedef struct {
        long tstamp;
        byte ra;
        byte rb;
} la_event;

la_event la_evbuf[0x10000];
unsigned short la_evp=0;
int la_nev=0;

void la_add_event(core *pcore)
{
        static int nested=0;
        unsigned short ix;

        if (nested) return;
        nested++;
        la_evbuf[la_evp].tstamp=pcore->Clocks;
        la_evbuf[la_evp].ra=(porta_rd(pcore) & pcore->trisa) |
                            (pcore->porta & (~pcore->trisa));
        la_evbuf[la_evp].rb=(portb_rd(pcore) & pcore->trisb) |
                            (pcore->portb & (~pcore->trisb));
        // Compress events (a change is needed to record an event)
        ix=la_evp-1;
        if (la_evbuf[la_evp].ra!=la_evbuf[ix].ra ||
            la_evbuf[la_evp].rb!=la_evbuf[ix].rb    ) {
                la_evp++;
                la_nev++; if (la_nev>0x10000) la_nev=0x10000;
        }
        nested--;
}

void la_show_events()
{
    int i,j,k,x,x0,x1,od,m,h;
    unsigned short ix;
    FILE *fp,*pp;

    x1=la_evbuf[la_evp].tstamp;
    ix=la_evp; ix-=la_nev-1;
    x0=la_evbuf[ix].tstamp;
    i=(x0+x1)/2;
    x0=i-10000; x1=i+10000;

    if ((pp=popen("gnuplot ","w"))==NULL) return;
    printf("\r\033[J\033m\033[1mCursors:\033[m center & zoom ");
    printf("\033[1mQ\033[muit "); fflush(stdout);
    fprintf(pp,"set yrange[0:29]\n");
    fprintf(pp,"set noytics\n");
    fprintf(pp,"set data style lines\n");
    for(;;) {
        fp=fopen("data.plt","w");
        // Port A Waveforms
        h=27;
        for (j=0;j<5;j++) {
            m=(1<<j);
            od=0;
            ix=la_evp; ix-=la_nev; k=0;
            for (i=0;i<la_nev;i++,ix++) {
                if (la_evbuf[ix].tstamp>=x0){
                    if(!k) fprintf(fp,"%d %d\n",x0,h+((od)?1:0));
                    k=1;
                    if (la_evbuf[ix].tstamp<=x1) {
                      fprintf(fp,"%d %d\n",la_evbuf[ix].tstamp,h+((od)?1:0));
                      od=la_evbuf[ix].ra & m;
                      fprintf(fp,"%d %d\n",la_evbuf[ix].tstamp,h+((od)?1:0));
                    }
                }
                if(!k) od=la_evbuf[ix].ra & m;
            }
            fprintf(fp,"%d %d\n",x1,h+((od)?1:0));
            h-=2;
            fprintf(fp,"\n");
        }
        // Port B Waveforms
        h-=2;
        for (j=0;j<8;j++) {
            m=(1<<j);
            od=0;
            ix=la_evp; ix-=la_nev; k=0;
            for (i=0;i<la_nev;i++,ix++) {
                if (la_evbuf[ix].tstamp>=x0){
                    if(!k) fprintf(fp,"%d %d\n",x0,h+((od)?1:0));
                    k=1;
                    if (la_evbuf[ix].tstamp<=x1) {
                      fprintf(fp,"%d %d\n",la_evbuf[ix].tstamp,h+((od)?1:0));
                      od=la_evbuf[ix].rb & m;
                      fprintf(fp,"%d %d\n",la_evbuf[ix].tstamp,h+((od)?1:0));
                    }
                }
                if(!k) od=la_evbuf[ix].rb & m;
            }
            fprintf(fp,"%d %d\n",x1,h+((od)?1:0));
            h-=2;
            fprintf(fp,"\n");
        }

        fclose(fp);
        fprintf(pp,"set xrange[%d:%d]\n",x0,x1);
        fprintf(pp,"plot \"data.plt\" t \"\"\n");
        fflush(pp);

        i=(x1-x0)/2;
        switch(getchar()){
        case 'q':
                pclose(pp);
                return;
        case 27: // Escaped cursor sequence
                if (getchar()!='[') break;
                switch(getchar()) {
                case 'A': // Zoom
                        x0+=i/2;
                        x1-=i/2;
                        break;
                case 'B': // Unzoom
                        x0-=i;
                        x1+=i;
                        break;
                case 'C': // Right
                        x0+=i/4;
                        x1+=i/4;
                        break;
                case 'D': // Left
                        x0-=i/4;
                        x1-=i/4;
                        break;
                }
        }
    }
}


