typedef unsigned char  byte;
typedef unsigned short word;

typedef struct {
        unsigned long Clocks;   // Clock counter for profiling
        unsigned int freq;      // Oscillator frequency/4
        word config;
        word ir;                // Prefetched Instruction. Not a visible reg
        word wdt;               // Watchdog counter
        word pc[8];
        byte sp;                // Not a visible reg
        byte prescaler;
        byte sync;
        byte sleep;
        byte w;
        byte tmr0;
        byte status;
        byte fsr;
        byte porta;
        byte portb;
        byte eedata;
        byte eedatawr;
        byte eeadr;
        byte pclatch;
        byte intcon;
        byte option;
        byte trisa;
        byte trisb;
        byte eecon1;
        byte eecon2;
        byte f[68];
        byte eeprom[64];
} core;

// Flags

#define CF   1
#define DCF  2
#define ZF   4
#define NPDF 8
#define NTOF 16
#define RP0F 32

enum { POR, WDTR };

// Prototypes
void f84_reset(core *pcore,int type);
void f84_run(core *pcore, word *prg,int nc);
char *f84_dis(word ireg);
void debug(core *pcore,word *prg);
byte porta_rd(core *pcore);
byte portb_rd(core *pcore);
void porta_wr(core *pcore);
void portb_wr(core *pcore);
void do_periodic_update();
void trace_interrupt(core *pcore);
void la_add_event(core *pcore);
void la_show_events();
void setcanon();
void setnocanon();

// Debug variables
enum {RUNNING,STEP,TRACE};
extern int debug_status;

